close all, clc, clear all;

%===================================================================================================%
%===================================================================================================%
%=====This code finds volumes in 3D data; designed for zebrafish adipocyte analysis=================%
%=====Miriam Moester (m.j.b.moester@vu.nl)==========================================================%
%=====Valentina Davidoiu (v.davidoiu@vu.nl)=========================================================%
%=====February 2017=================================================================================%
%=====MATLAB R2014a (8.3.0.532), 32-bit=============================================================%
%===================================================================================================%
%===================================================================================================%

home_folder=pwd;
addpath([home_folder filesep 'BM3D']);
parpool();

%=======Here are your folders with the raw images that you want to segment==========================%
data_folder=([home_folder filesep 'RAW DATA']);

%---- Input the name of the folder, containing the images--------%
%===== Input should be .tiff files, at 512x512 pixels, 8-bit grayscale with black background, white signal, z-stacks as separate files, no other files in data folder =======================%
data='FISH HFD 6 TIF'; %name of the fish that will be analysed
fprintf('Analysing dataset %s\n', data)

%======Here you need to give some input parameters=========
ps=0.4563; %this is PIXEL SIZE in um3
pixel_thr=100;% regions in 2D smaller than this value are discarded
rd=10; % after denoising your data the borders of the objects will be adjusted in order to be more uniform, for this you need to set up the radius of a disk that will fill the gaps, by default is 10

%===Folder location where your data will be saved after BM3D denoising and 3D watershed segmentation ====%
output_folder_path =([data_folder filesep data]);
savpath_bm3d=[output_folder_path filesep 'BM3D_denoised_images'];

if (exist(savpath_bm3d,'dir') ~= 7)
    mkdir(savpath_bm3d);
end

%  === Step 1: Load the images ===%
[Vol, img_info]=load_img_stack(output_folder_path,2);

VolSize = img_info.dims;
Vol_binary=zeros(VolSize);
Vol_binary_disk=zeros(VolSize);
Vol_denoised_bm3d=zeros(VolSize);

std_final=zeros(1,VolSize(3));
std_initial=zeros(1,VolSize(3));

% % ======Step 2: Determine threshold for binarisation, using OTSU method======%

totallev=zeros(1,VolSize(3));

for no=1:VolSize(3)
    lev=graythresh(Vol(:,:,no));
    totallev(no)=lev;
end

% ==== Discard the threshold information from the lowest 10% of values
min_thr=min(totallev(:));
max_thr=max(totallev(:));
back_thr=0.1*(max_thr-min_thr)+min_thr;
j=1;

for i=1: length(totallev)
    thr=totallev(i);
    
    if thr>back_thr
        newlev(j)=thr;
        j=j+1;
    end
    
end

average_level=mean(newlev(:));
% average_level=mean(totallev(:)); % bypass discarding lowest 10% thresholds with this option

%===== Step 3: Computing the average standard deviation, which is needed for BM3D===%

parfor no=1:VolSize(3)
    img=double(Vol(:,:,no));
    std_initial(1,no)=std2(img(:));   
end

sigma=round(mean(std_initial(:)));

% ===== Step 4: Denoising and binarisation of the images ====== %

parfor no=1:VolSize(3)   
    % =============rescaling the images between [0 1]===========================%
    img=double(Vol(:,:,no));
    maxzans = max(img(:));
    minzans = min(img(:));
    img_rescale= (img-minzans)/(maxzans-minzans);
    factor=(maxzans-minzans)+minzans;
    
    % =============Denoising the data via BM3D===========================%
    [~, BM] = BM3D(0, img_rescale, sigma);
    bm=factor*double(BM);
    std_final(1,no)=std2(bm(:));
    Vol_denoised_bm3d(:,:,no)=bm;
    
    % =============Saving the denoised images===========================%
    imwrite(uint8(bm),[savpath_bm3d filesep 'BM3D_sigma_' num2str(sigma) '_' num2str(no) '.bmp']);

    % =============Binarisation===========================%
    bw = im2bw(uint8(Vol_denoised_bm3d(:,:,no)),average_level); % with BM3D
%     bw = im2bw(Vol(:,:,no),average_level); % without BM3D
    Vol_binary (:,:,no)=bw;
    
    %========= Remove regions below the pixel threshold and filling small gaps==========%
    bwt = bwareaopen(bw,pixel_thr);
    se = strel('disk',rd);
    bwt = imclose(bwt,se);
    
    %========= Fill any holes, so that regionprops can be used to estimate the area enclosed by each of the boundaries
    bwh = imfill(bwt,'holes');
    Vol_binary_disk(:,:,no)=bwh;
   
end

fprintf('\t Stack of %2d images processed\n', VolSize(3))

%==== Step 5: Finding the connected components ===== %

% ===========Compute the distance transform to amplify the ridges between components =====%%%%%
D = bwdist(~Vol_binary_disk);

%========Complement the distance transform, force nonobject pixels to be -Inf, and then compute the watershed transform.
D = -D;
D(~Vol_binary_disk) = -Inf;

%=== Use watershed to find separate components ==== %
L = watershed(D);
W = L-1; %set background watershed component to 0 instead of 1

% == Find number of components and their centroids, used for making figure
CC = bwconncomp(W);
stats=regionprops(W,'Centroid');
centroids = cat(1, stats.Centroid);

% === Make a figure with all components ===== %
figure('Visible','Off'), axis square, axis([0 512 0 512]) % for images of 512x512 pixels
colormap(jet(CC.NumObjects+1)) % number of components plus one for easy visualisation
hold on
set(gca,'YDir','Reverse','color','none')

% plot all the components in the watershed matrix, with the background transparent
for s=1:VolSize(3)
    image(W(:,:,s),'AlphaData',double(W(:,:,s)))
end

% overlay the component number on the centroid of the component
for t=1:CC.NumObjects
    text(centroids(t,1),centroids(t,2),sprintf('%2.0f',t),'FontSize',15) 
end

cd(output_folder_path);
eval('print -dtiff all_components.tiff');
close all;

cd(home_folder);

% === Step 6: Calculating the volume of each component and output the information ==== %
vol_total= zeros(1,CC.NumObjects);

parfor k=1:CC.NumObjects

    [vol_out]=calCC_segmented(double(W==k),output_folder_path,k,ps,average_level,pixel_thr,data);
    vol_total(1,k)=vol_out;
end

% == Writing an analysis summary file == %
cd(output_folder_path)
fileID=fopen('all_volume_details.txt','wt');
fprintf(fileID,'%s\n',data);
fprintf(fileID,'\n');
fprintf(fileID,'%6s \n','The threshold used is:');
fprintf(fileID,'%6.4f\n',average_level);
fprintf(fileID,'\n');
fprintf(fileID,'%6s\n','For the given pixel size [cubic micrometers]:');
fprintf(fileID,'%6.4f\n',ps);
fprintf(fileID,'\n');
fprintf(fileID,'%4s\n','For the given pixel threshold:');
fprintf(fileID,'%6.0f\n',pixel_thr);
fprintf(fileID,'\n');
for k=1:length(CC.PixelIdxList)
    fprintf(fileID,'%6s',['The total volume of component ' num2str(k) ' is: ']);
    fprintf(fileID,'%9.2f\n',vol_total(1,k));
end
fclose(fileID);

cd(home_folder)

delete(gcp('nocreate')); 
clear all;

