clc; close all; clear all;
parpool();

%===================================================================================================%
%===================================================================================================%
%=====This code resegments components by 2D watershed; designed for zebrafish adipocyte analysis====%
%=====Miriam Moester (m.j.b.moester@vu.nl)==========================================================%
%=====Valentina Davidoiu (v.davidoiu@vu.nl)=========================================================%
%=====February 2017=================================================================================%
%=====MATLAB R2014a (8.3.0.532), 32-bit=============================================================%
%===================================================================================================%
%===================================================================================================%

home_folder=pwd;
addpath([home_folder filesep 'BM3D']);

%=======Here are your folders with the raw images that you want to segment==============================%
data_folder=([home_folder filesep 'RAW DATA']); % remove the all_components analysis image from this folder!
data='FISH TBT 3 TIF'; % name of the fish that will be analysed
output_folder_path =([data_folder filesep data]);

%========Select here the parts that need to be re-segmented ===%
comp=2; % input the component incorrectly segmented
slice_no=[8:11]; % input the slices incorrectly segmented

%======Here you need to give the same input parameters as in the original=======%
ps=0.4563; %this is PIXEL SIZE in um3
pixel_thr=100;% regions in 2D smaller than this value are discarded
rd=10; % after denoising your data the borders of the objects will be adjusted in order to be more uniform, for this you need to set up the radius of a disk that will fill the gaps, by default is 10

%  === Step 1: Load the images ===%
[Vol, img_info]=load_img_stack(output_folder_path,2);

VolSize = img_info.dims;
Vol_binary=zeros(VolSize);
Vol_binary_disk=zeros(VolSize);
Vol_denoised_bm3d=zeros(VolSize);

std_final=zeros(1,VolSize(3));
std_initial=zeros(1,VolSize(3));

% % ======Step 2: Determine threshold for binarisation, using OTSU method======%

totallev=zeros(1,VolSize(3));

for no=1:VolSize(3)
    lev=graythresh(Vol(:,:,no));
    totallev(no)=lev;
end

% ==== Discard the threshold information from the lowest 10% of values
min_thr=min(totallev(:));
max_thr=max(totallev(:));
back_thr=0.1*(max_thr-min_thr)+min_thr;
j=1;

for i=1: length(totallev)
    thr=totallev(i);
    
    if thr>back_thr
        newlev(j)=thr;
        j=j+1;
    end
    
end

average_level=mean(newlev(:));
% average_level=mean(totallev(:)); % bypass discarding lowest 10% thresholds with this option

%===== Step 3: Computing the average standard deviation, which is needed for BM3D===%

parfor no=1:VolSize(3)
    img=double(Vol(:,:,no));
    std_initial(1,no)=std2(img(:));   
end

sigma=round(mean(std_initial(:)));

% ===== Step 4: Denoising and binarisation of the images ====== %

parfor no=1:VolSize(3)   
    % =============rescaling the images between [0 1]===========================%
    img=double(Vol(:,:,no));
    maxzans = max(img(:));
    minzans = min(img(:));
    img_rescale= (img-minzans)/(maxzans-minzans);
    
    factor=(maxzans-minzans)+minzans;
    
    % =============Denoising the data via BM3D===========================%
    [~, BM] = BM3D(0, img_rescale, sigma);
    bm=factor*double(BM);
    std_final(1,no)=std2(bm(:));
    Vol_denoised_bm3d(:,:,no)=bm;
    
    % =============Binarisation===========================%
    bw = im2bw(uint8(Vol_denoised_bm3d(:,:,no)),average_level); % with BM3D
%     bw = im2bw(Vol(:,:,no),average_level); % without BM3D
    Vol_binary (:,:,no)=bw;
    
    %========= Remove regions below the pixel threshold and filling small gaps==========%
    bwt = bwareaopen(bw,pixel_thr);
    se = strel('disk',rd);
    bwt = imclose(bwt,se);
    
    %========= Fill any holes, so that regionprops can be used to estimate the area enclosed by each of the boundaries
    bwh = imfill(bwt,'holes');
    Vol_binary_disk(:,:,no)=bwh;
   
end

fprintf('\t Stack of %2d images processed\n', VolSize(3))

%==== Step 5: Finding the connected components ===== %
% ===========Compute the distance transform to amplify the ridges between components ===%
D = bwdist(~Vol_binary_disk);

%========Complement the distance transform, force nonobject pixels to be -Inf, and then compute the watershed transform.
D = -D;
D(~Vol_binary_disk) = -Inf;
%=== Use watershed to find separate components ==== %
L = watershed(D);
W = L-1; %set background watershed component to 0 instead of 1

% ==== Step 6: Doing watershed in 2D only (per slice) =====%

% step 1: creating empty matrixes with the correct size

vol_comp=zeros(VolSize(1),VolSize(2),length(slice_no));
D_comp=zeros(VolSize(1),VolSize(2),length(slice_no));
L_comp=zeros(VolSize(1),VolSize(2),length(slice_no));
CC_comp=zeros(1,length(slice_no));

% step 2: select desired component from watershed matrix
slice=1;
for no=slice_no(1):slice_no(end)
    vol_comp(:,:,slice)=double(W(:,:,no)==comp);
    slice=slice+1;
end
clear slice
% step 3: do 2D watershed

for no=1:length(slice_no)
        D_comp(:,:,no)=bwdist(~vol_comp(:,:,no));
end

D_comp=-D_comp;
D_comp(~vol_comp)=-Inf;

for no=1:length(slice_no)
    L_comp(:,:,no)=watershed(D_comp(:,:,no));
end

W_comp=L_comp-1;

% step 4: find number of components in new watershed per slice
for no=1:length(slice_no)
    CC_comp(1,no)=max(max(W_comp(:,:,no)));
end

% step 5: send this information into calCC_segmented version for 2D, per slice
for no=1:length(slice_no)

    calCC_segmented_for_2D(double(W_comp(:,:,no)),CC_comp(1,no),slice_no(no),output_folder_path,comp,ps,average_level,pixel_thr,data,home_folder);
    
end

delete(gcp);
